@@ -12,7 +12,7 @@ from __future__ import absolute_import
 #
 # Imports
 #
-
+import copy
 import errno
 import itertools
 import os
@@ -26,10 +26,12 @@ import warnings
 from collections import deque
 from functools import partial
 
-from . import Event, Process, cpu_count
+from . import cpu_count, get_context
 from . import util
-from .common import pickle_loads, reset_signals, restart_state
-from .compat import get_errno, send_offset
+from .common import (
+    TERM_SIGNAL, human_status, pickle_loads, reset_signals, restart_state,
+)
+from .compat import get_errno, mem_rss, send_offset
 from .einfo import ExceptionInfo
 from .dummy import DummyProcess
 from .exceptions import (
@@ -44,6 +46,10 @@ from .exceptions import (
 from .five import Empty, Queue, range, values, reraise, monotonic
 from .util import Finalize, debug
 
+MAXMEM_USED_FMT = """\
+child process exiting after exceeding memory limit ({0}KiB / {1}KiB)
+"""
+
 PY3 = sys.version_info[0] == 3
 
 if platform.system() == 'Windows':  # pragma: no cover
@@ -51,8 +57,10 @@ if platform.system() == 'Windows':  # pragma: no cover
     # handled by # any process, so this is needed to terminate the task
     # *and its children* (if any).
     from ._win import kill_processtree as _kill  # noqa
+    SIGKILL = TERM_SIGNAL
 else:
     from os import kill as _kill                 # noqa
+    SIGKILL = signal.SIGKILL
 
 
 try:
@@ -67,10 +75,6 @@ else:
     # Semaphore is a factory function pointing to _Semaphore
     _Semaphore = threading._Semaphore  # noqa
 
-SIGMAP = dict(
-    (getattr(signal, n), n) for n in dir(signal) if n.startswith('SIG')
-)
-
 #
 # Constants representing the state of a pool
 #
@@ -122,15 +126,6 @@ def _get_send_offset(connection):
     return native
 
 
-def human_status(status):
-    if status or 0 < 0:
-        try:
-            return 'signal {0} ({1})'.format(-status, SIGMAP[-status])
-        except KeyError:
-            return 'signal {0}'.format(-status)
-    return 'exitcode {0}'.format(status)
-
-
 def mapstar(args):
     return list(map(*args))
 
@@ -140,12 +135,11 @@ def starmapstar(args):
 
 
 def error(msg, *args, **kwargs):
-    if util._logger:
-        util._logger.error(msg, *args, **kwargs)
+    util.get_logger().error(msg, *args, **kwargs)
 
 
 def stop_if_not_current(thread, timeout=None):
-    if thread is not threading.currentThread():
+    if thread is not threading.current_thread():
         thread.stop(timeout)
 
 
@@ -153,29 +147,22 @@ class LaxBoundedSemaphore(_Semaphore):
     """Semaphore that checks that # release is <= # acquires,
     but ignores if # releases >= value."""
 
-    def __init__(self, value=1, verbose=None):
-        if PY3:
-            _Semaphore.__init__(self, value)
-        else:
-            _Semaphore.__init__(self, value, verbose)
-        self._initial_value = value
-
-    def grow(self):
-        if PY3:
-            cond = self._cond
-        else:
-            cond = self._Semaphore__cond
-        with cond:
-            self._initial_value += 1
-            self._Semaphore__value += 1
-            cond.notify()
-
     def shrink(self):
         self._initial_value -= 1
         self.acquire()
 
     if PY3:
 
+        def __init__(self, value=1, verbose=None):
+            _Semaphore.__init__(self, value)
+            self._initial_value = value
+
+        def grow(self):
+            with self._cond:
+                self._initial_value += 1
+                self._value += 1
+                self._cond.notify()
+
         def release(self):
             cond = self._cond
             with cond:
@@ -188,6 +175,17 @@ class LaxBoundedSemaphore(_Semaphore):
                 _Semaphore.release(self)
     else:
 
+        def __init__(self, value=1, verbose=None):
+            _Semaphore.__init__(self, value, verbose)
+            self._initial_value = value
+
+        def grow(self):
+            cond = self._Semaphore__cond
+            with cond:
+                self._initial_value += 1
+                self._Semaphore__value += 1
+                cond.notify()
+
         def release(self):  # noqa
             cond = self._Semaphore__cond
             with cond:
@@ -214,7 +212,7 @@ class MaybeEncodingError(Exception):
         super(MaybeEncodingError, self).__init__(self.exc, self.value)
 
     def __repr__(self):
-        return "<MaybeEncodingError: %s>" % str(self)
+        return "<%s: %s>" % (self.__class__.__name__, str(self))
 
     def __str__(self):
         return "Error sending result: '%r'. Reason: '%r'." % (
@@ -233,45 +231,47 @@ def soft_timeout_sighandler(signum, frame):
 #
 
 
-class Worker(Process):
-    _controlled_termination = False
-    _job_terminated = False
+class Worker(object):
 
     def __init__(self, inq, outq, synq=None, initializer=None, initargs=(),
                  maxtasks=None, sentinel=None, on_exit=None,
-                 sigprotection=True):
+                 sigprotection=True, wrap_exception=True,
+                 max_memory_per_child=None):
         assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
         self.initializer = initializer
         self.initargs = initargs
         self.maxtasks = maxtasks
+        self.max_memory_per_child = max_memory_per_child
         self._shutdown = sentinel
         self.on_exit = on_exit
         self.sigprotection = sigprotection
         self.inq, self.outq, self.synq = inq, outq, synq
-        self._make_shortcuts()
+        self.wrap_exception = wrap_exception  # XXX cannot disable yet
+        self.contribute_to_object(self)
 
-        super(Worker, self).__init__()
+    def contribute_to_object(self, obj):
+        obj.inq, obj.outq, obj.synq = self.inq, self.outq, self.synq
+        obj.inqW_fd = self.inq._writer.fileno()    # inqueue write fd
+        obj.outqR_fd = self.outq._reader.fileno()  # outqueue read fd
+        if self.synq:
+            obj.synqR_fd = self.synq._reader.fileno()  # synqueue read fd
+            obj.synqW_fd = self.synq._writer.fileno()  # synqueue write fd
+            obj.send_syn_offset = _get_send_offset(self.synq._writer)
+        else:
+            obj.synqR_fd = obj.synqW_fd = obj._send_syn_offset = None
+        obj._quick_put = self.inq._writer.send
+        obj._quick_get = self.outq._reader.recv
+        obj.send_job_offset = _get_send_offset(self.inq._writer)
+        return obj
 
     def __reduce__(self):
         return self.__class__, (
             self.inq, self.outq, self.synq, self.initializer,
-            self.initargs, self.maxtasks, self._shutdown,
+            self.initargs, self.maxtasks, self._shutdown, self.on_exit,
+            self.sigprotection, self.wrap_exception,
         )
 
-    def _make_shortcuts(self):
-        self.inqW_fd = self.inq._writer.fileno()    # inqueue write fd
-        self.outqR_fd = self.outq._reader.fileno()  # outqueue read fd
-        if self.synq:
-            self.synqR_fd = self.synq._reader.fileno()  # synqueue read fd
-            self.synqW_fd = self.synq._writer.fileno()  # synqueue write fd
-            self.send_syn_offset = _get_send_offset(self.synq._writer)
-        else:
-            self.synqR_fd = self.synqW_fd = self._send_syn_offset = None
-        self._quick_put = self.inq._writer.send
-        self._quick_get = self.outq._reader.recv
-        self.send_job_offset = _get_send_offset(self.inq._writer)
-
-    def run(self):
+    def __call__(self):
         _exit = sys.exit
         _exitcode = [None]
 
@@ -312,10 +312,6 @@ class Worker(Process):
     def on_loop_start(self, pid):
         pass
 
-    def terminate_controlled(self):
-        self._controlled_termination = True
-        self.terminate()
-
     def prepare_result(self, result):
         return result
 
@@ -325,6 +321,7 @@ class Worker(Process):
         inqW_fd = self.inqW_fd
         synqW_fd = self.synqW_fd
         maxtasks = self.maxtasks
+        max_memory_per_child = self.max_memory_per_child or 0
         prepare_result = self.prepare_result
 
         wait_for_job = self.wait_for_job
@@ -374,6 +371,15 @@ class Worker(Process):
                     finally:
                         del(tb)
                 completed += 1
+                if max_memory_per_child > 0:
+                    used_kb = mem_rss()
+                    if used_kb <= 0:
+                        error('worker unable to determine memory usage')
+                    if used_kb > 0 and used_kb > max_memory_per_child:
+                        error(MAXMEM_USED_FMT.format(
+                            used_kb, max_memory_per_child))
+                        return EX_RECYCLE
+
         debug('worker exiting after %d tasks', completed)
         if maxtasks:
             return EX_RECYCLE if completed == maxtasks else EX_FAILURE
@@ -473,7 +479,7 @@ class PoolThread(DummyProcess):
         except RestartFreqExceeded as exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
                   exc_info=1)
-            _kill(os.getpid(), signal.SIGTERM)
+            _kill(os.getpid(), TERM_SIGNAL)
             sys.exit()
         except Exception as exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
@@ -539,20 +545,23 @@ class Supervisor(PoolThread):
 
 class TaskHandler(PoolThread):
 
-    def __init__(self, taskqueue, put, outqueue, pool):
+    def __init__(self, taskqueue, put, outqueue, pool, cache):
         self.taskqueue = taskqueue
         self.put = put
         self.outqueue = outqueue
         self.pool = pool
+        self.cache = cache
         super(TaskHandler, self).__init__()
 
     def body(self):
+        cache = self.cache
         taskqueue = self.taskqueue
         put = self.put
 
         for taskseq, set_length in iter(taskqueue.get, None):
+            task = None
+            i = -1
             try:
-                i = -1
                 for i, task in enumerate(taskseq):
                     if self._state:
                         debug('task handler found thread._state != RUN')
@@ -562,15 +571,25 @@ class TaskHandler(PoolThread):
                     except IOError:
                         debug('could not put task on queue')
                         break
+                    except Exception:
+                        job, ind = task[:2]
+                        try:
+                            cache[job]._set(ind, (False, ExceptionInfo()))
+                        except KeyError:
+                            pass
                 else:
                     if set_length:
                         debug('doing set_length()')
                         set_length(i + 1)
                     continue
                 break
-            except Exception as exc:
-                error('Task Handler ERROR: %r', exc, exc_info=1)
-                break
+            except Exception:
+                job, ind = task[:2] if task else (0, 0)
+                if job in cache:
+                    cache[job]._set(ind + 1, (False, ExceptionInfo()))
+                if set_length:
+                    util.debug('doing set_length()')
+                    set_length(i + 1)
         else:
             debug('task handler got sentinel')
 
@@ -622,8 +641,7 @@ class TimeoutHandler(PoolThread):
             return
 
         # Run timeout callback
-        if job._timeout_callback is not None:
-            job._timeout_callback(soft=True, timeout=job._soft_timeout)
+        job.handle_timeout(soft=True)
 
         try:
             _kill(job._worker_pid, SIG_SOFT_TIMEOUT)
@@ -647,15 +665,19 @@ class TimeoutHandler(PoolThread):
         process, _index = self._process_by_pid(job._worker_pid)
 
         # Run timeout callback
-        if job._timeout_callback is not None:
-            job._timeout_callback(soft=False, timeout=job._timeout)
+        job.handle_timeout(soft=False)
+
         if process:
             self._trywaitkill(process)
 
     def _trywaitkill(self, worker):
         debug('timeout: sending TERM to %s', worker._name)
         try:
-            worker.terminate()
+            if os.getpgid(worker.pid) == worker.pid:
+                debug("worker %s is a group leader. It is safe to kill (SIGTERM) the whole group", worker.pid)
+                os.killpg(os.getpgid(worker.pid), signal.SIGTERM)
+            else:
+                worker.terminate()
         except OSError:
             pass
         else:
@@ -663,12 +685,16 @@ class TimeoutHandler(PoolThread):
                 return
         debug('timeout: TERM timed-out, now sending KILL to %s', worker._name)
         try:
-            _kill(worker.pid, signal.SIGKILL)
+            if os.getpgid(worker.pid) == worker.pid:
+                debug("worker %s is a group leader. It is safe to kill (SIGKILL) the whole group", worker.pid)
+                os.killpg(os.getpgid(worker.pid), signal.SIGKILL)
+            else:
+                _kill(worker.pid, SIGKILL)
         except OSError:
             pass
 
     def handle_timeouts(self):
-        cache = self.cache
+        cache = copy.deepcopy(self.cache)
         t_hard, t_soft = self.t_hard, self.t_soft
         dirty = set()
         on_soft_timeout = self.on_soft_timeout
@@ -774,7 +800,7 @@ class ResultHandler(PoolThread):
 
         def on_death(pid, exitcode):
             try:
-                os.kill(pid, signal.SIGTERM)
+                os.kill(pid, TERM_SIGNAL)
             except OSError as exc:
                 if get_errno(exc) != errno.ESRCH:
                     raise
@@ -900,6 +926,7 @@ class Pool(object):
     '''
     Class which supports an async version of applying functions to arguments.
     '''
+    _wrap_exception = True
     Worker = Worker
     Supervisor = Supervisor
     TaskHandler = TaskHandler
@@ -921,7 +948,11 @@ class Pool(object):
                  allow_restart=False,
                  synack=False,
                  on_process_exit=None,
+                 context=None,
+                 max_memory_per_child=None,
+                 enable_timeouts=False,
                  **kwargs):
+        self._ctx = context or get_context()
         self.synack = synack
         self._setup_queues()
         self._taskqueue = Queue()
@@ -930,6 +961,7 @@ class Pool(object):
         self.timeout = timeout
         self.soft_timeout = soft_timeout
         self._maxtasksperchild = maxtasksperchild
+        self._max_memory_per_child = max_memory_per_child
         self._initializer = initializer
         self._initargs = initargs
         self._on_process_exit = on_process_exit
@@ -942,6 +974,12 @@ class Pool(object):
         self.readers = {}
         self.allow_restart = allow_restart
 
+        self.enable_timeouts = bool(
+            enable_timeouts or
+            self.timeout is not None or
+            self.soft_timeout is not None
+        )
+
         if soft_timeout and SIG_SOFT_TIMEOUT is None:
             warnings.warn(UserWarning(
                 "Soft timeouts are not supported: "
@@ -959,6 +997,8 @@ class Pool(object):
         if on_process_exit is not None and not callable(on_process_exit):
             raise TypeError('on_process_exit must be callable')
 
+        self._Process = self._ctx.Process
+
         self._pool = []
         self._poolctrl = {}
         self.putlocks = putlocks
@@ -973,25 +1013,30 @@ class Pool(object):
         self._task_handler = self.TaskHandler(self._taskqueue,
                                               self._quick_put,
                                               self._outqueue,
-                                              self._pool)
+                                              self._pool,
+                                              self._cache)
         if threads:
             self._task_handler.start()
 
-        # Thread killing timedout jobs.
-        self._timeout_handler = self.TimeoutHandler(
-            self._pool, self._cache,
-            self.soft_timeout, self.timeout,
-        )
-        self._timeout_handler_mutex = Lock()
-        self._timeout_handler_started = False
-        if self.timeout is not None or self.soft_timeout is not None:
-            self._start_timeout_handler()
-
-        # If running without threads, we need to check for timeouts
-        # while waiting for unfinished work at shutdown.
         self.check_timeouts = None
-        if not threads:
-            self.check_timeouts = self._timeout_handler.handle_event
+
+        # Thread killing timedout jobs.
+        if self.enable_timeouts:
+            self._timeout_handler = self.TimeoutHandler(
+                self._pool, self._cache,
+                self.soft_timeout, self.timeout,
+            )
+            self._timeout_handler_mutex = Lock()
+            self._timeout_handler_started = False
+            self._start_timeout_handler()
+            # If running without threads, we need to check for timeouts
+            # while waiting for unfinished work at shutdown.
+            if not threads:
+                self.check_timeouts = self._timeout_handler.handle_event
+        else:
+            self._timeout_handler = None
+            self._timeout_handler_started = False
+            self._timeout_handler_mutex = None
 
         # Thread processing results in the outqueue.
         self._result_handler = self.create_result_handler()
@@ -1010,6 +1055,12 @@ class Pool(object):
             exitpriority=15,
         )
 
+    def Process(self, *args, **kwds):
+        return self._Process(*args, **kwds)
+
+    def WorkerProcess(self, worker):
+        return worker.contribute_to_object(self.Process(target=worker))
+
     def create_result_handler(self, **extra_kwargs):
         return self.ResultHandler(
             self._outqueue, self._quick_get, self._cache,
@@ -1046,15 +1097,17 @@ class Pool(object):
         return self._inqueue, self._outqueue, None
 
     def _create_worker_process(self, i):
-        sentinel = Event() if self.allow_restart else None
+        sentinel = self._ctx.Event() if self.allow_restart else None
         inq, outq, synq = self.get_process_queues()
-        w = self.Worker(
+        w = self.WorkerProcess(self.Worker(
             inq, outq, synq, self._initializer, self._initargs,
             self._maxtasksperchild, sentinel, self._on_process_exit,
             # Need to handle all signals if using the ipc semaphore,
             # to make sure the semaphore is released.
             sigprotection=self.threads,
-        )
+            wrap_exception=self._wrap_exception,
+            max_memory_per_child=self._max_memory_per_child,
+        ))
         self._pool.append(w)
         self._process_register_queues(w, (inq, outq, synq))
         w.name = w.name.replace('Process', 'PoolWorker')
@@ -1209,7 +1262,6 @@ class Pool(object):
         for worker in self._pool:
             if not self._worker_active(worker):
                 yield worker
-        raise StopIteration()
 
     def _worker_active(self, worker):
         for job in values(self._cache):
@@ -1265,9 +1317,8 @@ class Pool(object):
                 raise
 
     def _setup_queues(self):
-        from billiard.queues import SimpleQueue
-        self._inqueue = SimpleQueue()
-        self._outqueue = SimpleQueue()
+        self._inqueue = self._ctx.SimpleQueue()
+        self._outqueue = self._ctx.SimpleQueue()
         self._quick_put = self._inqueue._writer.send
         self._quick_get = self._outqueue._reader.recv
 
@@ -1280,7 +1331,7 @@ class Pool(object):
     def _start_timeout_handler(self):
         # ensure more than one thread does not start the timeout handler
         # thread at once.
-        if self.threads:
+        if self.threads and self._timeout_handler is not None:
             with self._timeout_handler_mutex:
                 if not self._timeout_handler_started:
                     self._timeout_handler_started = True
@@ -1442,7 +1493,7 @@ class Pool(object):
         proc, _ = self._process_by_pid(pid)
         if proc is not None:
             try:
-                _kill(pid, sig or signal.SIGTERM)
+                _kill(pid, sig or TERM_SIGNAL)
             except OSError as exc:
                 if get_errno(exc) != errno.ESRCH:
                     raise
@@ -1527,7 +1578,7 @@ class Pool(object):
         stop_if_not_current(self._result_handler)
         debug('result handler joined')
         for i, p in enumerate(self._pool):
-            debug('joining worker %s/%s (%r)', i+1, len(self._pool), p)
+            debug('joining worker %s/%s (%r)', i + 1, len(self._pool), p)
             if p._popen is not None:  # process started?
                 p.join()
         debug('pool join complete')
@@ -1629,7 +1680,6 @@ class ApplyResult(object):
         self._error_callback = error_callback
         self._timeout_callback = timeout_callback
         self._timeout = timeout
-        self._terminated = None
         self._soft_timeout = soft_timeout
         self._lost_worker_timeout = lost_worker_timeout
         self._on_timeout_set = on_timeout_set
@@ -1641,10 +1691,12 @@ class ApplyResult(object):
         self._cancelled = False
         self._worker_pid = None
         self._time_accepted = None
+        self._terminated = None
         cache[self._job] = self
 
     def __repr__(self):
-        return '<Result: {id} ack:{ack} ready:{ready}>'.format(
+        return '<%s: {id} ack:{ack} ready:{ready}>'.format(
+            self.__class__.__name__,
             id=self._job, ack=self._accepted, ready=self.ready(),
         )
 
@@ -1689,16 +1741,23 @@ class ApplyResult(object):
         else:
             raise self._value.exception
 
-    def safe_apply_callback(self, fun, *args):
+    def safe_apply_callback(self, fun, *args, **kwargs):
         if fun:
             try:
-                fun(*args)
+                fun(*args, **kwargs)
             except self._callbacks_propagate:
                 raise
             except Exception as exc:
                 error('Pool callback raised exception: %r', exc,
                       exc_info=1)
 
+    def handle_timeout(self, soft=False):
+        if self._timeout_callback is not None:
+            self.safe_apply_callback(
+                self._timeout_callback, soft=soft,
+                timeout=self._soft_timeout if soft else self._timeout,
+            )
+
     def _set(self, i, obj):
         with self._mutex:
             if self._on_timeout_cancel:
@@ -1798,9 +1857,9 @@ class MapResult(ApplyResult):
                 self._cache.pop(self._job, None)
             self._event.set()
 
-    def _ack(self, i, time_accepted, pid):
+    def _ack(self, i, time_accepted, pid, *args):
         start = i * self._chunksize
-        stop = (i + 1) * self._chunksize
+        stop = min((i + 1) * self._chunksize, self._length)
         for j in range(start, stop):
             self._accepted[j] = True
             self._worker_pid[j] = pid
@@ -1887,7 +1946,7 @@ class IMapIterator(object):
                 self._cond.notify()
                 del self._cache[self._job]
 
-    def _ack(self, i, time_accepted, pid):
+    def _ack(self, i, time_accepted, pid, *args):
         self._worker_pids.append(pid)
 
     def ready(self):
@@ -1919,7 +1978,7 @@ class IMapUnorderedIterator(IMapIterator):
 
 class ThreadPool(Pool):
 
-    from billiard.dummy import Process as DummyProcess
+    from .dummy import Process as DummyProcess
     Process = DummyProcess
 
     def __init__(self, processes=None, initializer=None, initargs=()):
